package de.uni_passau.fim.seibt.gitwrapper.repo;

import de.uni_passau.fim.seibt.gitwrapper.process.ProcessExecutor.ExecRes;

import java.time.Instant;
import java.time.OffsetDateTime;
import java.time.ZoneId;
import java.util.*;
import java.util.function.Function;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

/**
 * A {@link Commit} made in a {@link Repository}.
 */
public class Commit {

    private static final Logger LOG = Logger.getLogger(Commit.class.getCanonicalName());

    // since there is no porcelain format for this command, this regex is might depend on the git version
    private static final Pattern TREE_INFO = Pattern.compile("tree (.{40})\\nparent (.{40})\\n");
    private static final Pattern AUTHOR_INFO = Pattern.compile("author (.*?)<(.*?)> (\\d+) ([+-]\\d{4})\\n");
    private static final Pattern COMMITTER_INFO = Pattern.compile("committer (.*?)<(.*?)> (\\d+) ([+-]\\d{4})\\n");

    protected GitWrapper git;
    protected Repository repo;

    private String id;

    private String message;

    private String author;
    private String authorMail;
    private OffsetDateTime authorTime;

    private String committer;
    private String committerMail;
    private OffsetDateTime committerTime;

    /**
     * Constructs a new {@link Commit} with the given <code>id</code> made in the <code>repo</code>.
     *
     * @param repo the {@link Repository} the {@link Commit} was made in
     * @param id   the ID of the {@link Commit}
     */
    Commit(Repository repo, String id) {
        this.git = repo.getGit();
        this.repo = repo;
        this.id = id;
    }

    /**
     * Returns the parents of this {@link Commit}.
     *
     * @return the parent {@link Commit} objects
     */
    public List<Commit> getParents() {
        Optional<ExecRes> revList = git.exec(repo.getDir(), "rev-list", "--parents", "-n", "1", id);
        Function<ExecRes, List<Commit>> toParentsList = res -> {
            String[] ids = res.stdOut.split("\\s+");

            LOG.fine(() -> String.format("Found %d parents for %s.", ids.length - 1, this));
            LOG.finer(() -> String.format("Commit id and parents are:%n%s", String.join(System.lineSeparator(), ids)));

            return Arrays.stream(ids).skip(1).map(repo::getCommitUnchecked).collect(Collectors.toList());
        };

        return revList.map(toParentsList).orElse(Collections.emptyList());
    }

    /**
     * Optionally returns the merge base for <code>this</code> and <code>other</code>. The <code>other</code> commit
     * must be part of the same {@link Repository} this {@link Commit} is.
     *
     * @param other the other {@link Commit}
     * @return the merge base or an empty {@link Optional} if there is no merge base or an exception occurred
     */
    public Optional<Commit> getMergeBase(Commit other) {

        if (!repo.equals(other.repo)) {
            LOG.warning(() -> {
                String msg = "Failed to obtain a merge base for %s and %s as they are not from the same repository.";
                return String.format(msg, this, other);
            });

            return Optional.empty();
        }

        Optional<ExecRes> mergeBase = git.exec(repo.getDir(), "merge-base", getId(), other.getId());
        Function<ExecRes, Commit> toCommit = res -> {

            if (git.failed(res)) {
                LOG.warning(() -> String.format("Failed to obtain a merge base for %s and %s.", this, other));
                return null;
            }

            Commit base = repo.getCommitUnchecked(res.stdOut);

            LOG.fine(() -> String.format("Commits %s and %s have the merge base %s.", this, other, base));

            return base;
        };

        return mergeBase.map(toCommit);
    }

    /**
     * Returns the ID of this commit.
     *
     * @return the ID
     */
    public String getId() {
        return id;
    }

    /**
     * Returns the author of this commit.
     *
     * @return the author
     */
    public String getAuthor() {
        if (author == null) {
            getCommitInfo();
        }

        return author;
    }

    /**
     * Sets the author to the new value.
     *
     * @param author the new author
     */
    void setAuthor(String author) {
        this.author = author;
    }

    /**
     * Returns the author mail of this commit.
     *
     * @return the author mail
     */
    public String getAuthorMail() {
        if (authorMail == null) {
            getCommitInfo();
        }

        return authorMail;
    }

    /**
     * Sets the author mail to the new value.
     *
     * @param authorMail the new author mail
     */
    void setAuthorMail(String authorMail) {
        this.authorMail = authorMail;
    }

    /**
     * Returns the author time of this commit.
     *
     * @return the author time
     */
    public OffsetDateTime getAuthorTime() {
        if (authorTime == null) {
            getCommitInfo();
        }

        return authorTime;
    }

    /**
     * Sets the author time to the new value.
     *
     * @param authorTime the new author time
     */
    void setAuthorTime(OffsetDateTime authorTime) {
        this.authorTime = authorTime;
    }

    /**
     * Returns the committer.
     *
     * @return the committer
     */
    public String getCommitter() {
        return committer;
    }

    /**
     * Sets the committer to the new value.
     *
     * @param committer the new committer
     */
    void setCommitter(String committer) {
        this.committer = committer;
    }

    /**
     * Returns the committer mail.
     *
     * @return the committer mail
     */
    public String getCommitterMail() {
        return committerMail;
    }

    /**
     * Sets the committer mail to the new value.
     *
     * @param committerMail the new committer mail
     */
    void setCommitterMail(String committerMail) {
        this.committerMail = committerMail;
    }

    /**
     * Returns the committer time.
     *
     * @return the committer time
     */
    public OffsetDateTime getCommitterTime() {
        return committerTime;
    }

    /**
     * Sets the committer time to the new value.
     *
     * @param committerTime the new committer time
     */
    void setCommitterTime(OffsetDateTime committerTime) {
        this.committerTime = committerTime;
    }

    /**
     * Returns the commit message.
     *
     * @return the commit message
     */
    public String getMessage() {
        return message;
    }

    /**
     * Sets the commit message.
     *
     * @param message the commit message
     */
    void setMessage(String message) {
        this.message = message;
    }

    /**
     * Reads the author information (author, author mail, author time) about this commit and saves it in the
     * {@link Commit}.
     */
    private void getCommitInfo() {
        git.exec(repo.getDir(), "cat-file", "-p", id).ifPresent(execRes -> {
            if (git.failed(execRes)) {
                LOG.warning(() -> String.format("Failed to obtain information about commit %s.", this));
                return;
            }

            String result = execRes.stdOut;
            Matcher matcher = AUTHOR_INFO.matcher(result);
            if (!matcher.find()) {
                LOG.warning(() -> String.format("Unexpected output while getting commit info for %s.", this));
            } else {
                author = matcher.group(1);
                authorMail = matcher.group(2);
                authorTime = OffsetDateTime.ofInstant(Instant.ofEpochSecond(Long.parseLong(matcher.group(3))), ZoneId.of(matcher.group(4)));
            }

            matcher = COMMITTER_INFO.matcher(result);
            if (!matcher.find()) {
                LOG.warning(() -> String.format("Unexpected output while getting commit info for %s.", this));
            } else {
                committer = matcher.group(1);
                committerMail = matcher.group(2);
                committerTime = OffsetDateTime.ofInstant(Instant.ofEpochSecond(Long.parseLong(matcher.group(3))), ZoneId.of(matcher.group(4)));
            }

            message = result.substring(result.lastIndexOf("\n"));
        });
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }

        if (o == null || getClass() != o.getClass()) {
            return false;
        }

        Commit commit = (Commit) o;
        return Objects.equals(repo, commit.repo) && Objects.equals(id, commit.id);
    }

    @Override
    public int hashCode() {
        return Objects.hash(repo, id);
    }

    @Override
    public String toString() {
        return String.valueOf(id);
    }
}
