package de.uni_passau.fim.seibt.gitwrapper.repo;

import de.uni_passau.fim.seibt.gitwrapper.process.ProcessExecutor.ExecRes;
import de.uni_passau.fim.seibt.gitwrapper.util.Pair;
import org.apache.commons.io.FileUtils;

import java.io.File;
import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.time.Instant;
import java.time.OffsetDateTime;
import java.time.ZoneId;
import java.util.*;
import java.util.function.Function;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Collectors;

/**
 * A git {@link Repository}.
 */
public class Repository {

    private static final Logger LOG = Logger.getLogger(Repository.class.getCanonicalName());

    // Constants defining expected commit header fields when parsing the output of 'git blame --porcelain'.
    private static final String AUTHOR = "author";
    private static final String AUTHOR_MAIL = "author-mail";
    private static final String AUTHOR_TIME = "author-time";
    private static final String AUTHOR_TZ = "author-tz";
    private static final String COMMITTER = "committer";
    private static final String COMMITTER_MAIL = "committer-mail";
    private static final String COMMITTER_TIME = "committer-time";
    private static final String COMMITTER_TZ = "committer-tz";
    private static final String FILENAME = "filename";

    // The type returned for commits by 'git cat-file -t".
    private static final String TYPE_COMMIT = "commit";

    private static final String CONFLICT_START = "<<<<<<<";
    private static final String CONFLICT_MIDDLE = "=======";
    private static final String CONFLICT_END = ">>>>>>>";

    private GitWrapper git;

    private String url;
    private File dir;

    private Map<String, Commit> commits;

    /**
     * Constructs a new {@link Repository}.
     *
     * @param git
     *         the {@link GitWrapper} to use
     * @param url
     *         the url the {@link Repository} was cloned from
     * @param dir
     *         the directory containing the {@link Repository}
     */
    Repository(GitWrapper git, String url, File dir) {
        this.git = git;
        this.url = url;

        try {
            this.dir = dir.getCanonicalFile();
        } catch (IOException e) {
            LOG.log(Level.WARNING, e, () -> "Exception computing the canonical file for " + dir + ". Using the absolute file instead.");
            this.dir = dir.getAbsoluteFile();
        }

        this.commits = new HashMap<>();
        this.commits.put(DummyCommit.DUMMY_COMMIT_ID, new DummyCommit(this));
    }

    /**
     * Performs a checkout of the given {@link Commit}.
     *
     * @param c
     *         the {@link Commit} to checkout
     * @return whether the checkout was successful
     * @see <a href=https://git-scm.com/docs/git-checkout>git checkout</a>
     */
    public boolean checkout(Commit c) {
        Optional<ExecRes> checkout = git.exec(dir, "checkout", c.getId());
        Function<ExecRes, Boolean> toBoolean = res -> {
            boolean failed = git.failed(res);

            if (failed) {
                LOG.warning(() -> String.format("Checkout of %s failed.", c));
            }

            return !failed;
        };

        return checkout.map(toBoolean).orElse(false);
    }

    /**
     * Performs a fetch in this {@link Repository}.
     *
     * @return whether the fetch was successful
     * @see <a href=https://git-scm.com/docs/git-fetch>git fetch</a>
     */
    public boolean fetch() {
        Optional<ExecRes> fetch = git.exec(dir, "fetch", "--all", "--tags"); //TODO add "--prune"?
        Function<ExecRes, Boolean> toBoolean = res -> {
            boolean failed = git.failed(res);

            if (failed) {
                LOG.warning(() -> String.format("Fetch of %s failed.", this));
            }

            return !failed;
        };

        return fetch.map(toBoolean).orElse(false);
    }

    /**
     * Returns the merge commits in this repository.
     *
     * @return the merge commits
     */
    public List<Commit> getMergeCommits() {
        Optional<ExecRes> revList = git.exec(dir, "rev-list", "--all", "--merges");
        Function<ExecRes, List<Commit>> toCommitList = res -> {

            if (git.failed(res)) {
                LOG.warning(() -> String.format("Failed to obtain the merge commits from %s.", this));
                return null;
            }

            String[] lines = res.stdOut.isEmpty() ? new String[] {} : res.stdOut.split("[\\r?\\n]+");

            LOG.fine(() -> String.format("Found %d merge commits in %s.", lines.length, this));

            if (lines.length > 0) {
                LOG.finer(() -> String.format("Merge commits are:%n%s", String.join(System.lineSeparator(), lines)));
            }

            return Arrays.stream(lines).map(this::getCommitUnchecked).collect(Collectors.toList());
        };

        return revList.map(toCommitList).orElse(Collections.emptyList());
    }

    /**
     * Returns the {@link Commit} currently pointed at by <code>HEAD</code>.
     *
     * @return the <code>HEAD</code> commit
     */
    public Optional<Commit> getCurrentHEAD() {
        Optional<ExecRes> revParse = git.exec(dir, "rev-parse", "HEAD");
        Function<ExecRes, Commit> toHEAD = res -> {

            if (git.failed(res)) {
                LOG.warning(() -> "Failed to obtain the current HEAD commit.");
                return null;
            }

            return getCommitUnchecked(res.stdOut);
        };

        return revParse.map(toHEAD);
    }

    /**
     * Returns a {@link Commit} for the given ID. The caller must ensure that the ID is a full SHA1 hash of a
     * commit that exists in this repository.
     *
     * @param id
     *         the ID for the {@link Commit}
     * @return the {@link Commit}
     */
    Commit getCommitUnchecked(String id) {
        return commits.computeIfAbsent(id, theID -> new Commit(this, theID));
    }

    /**
     * Returns a {@link Commit} for the given ID. If the given ID does not designate a commit that exists in this
     * {@link Repository} an empty {@link Optional} will be returned. The ID will be resolved to a full SHA1 hash.
     *
     * @param id
     *         the ID of the commit
     * @return the {@link Commit} or an empty {@link Optional} if the ID is invalid or an exception occurs
     */
    public Optional<Commit> getCommit(String id) {

        if (commits.containsKey(id)) {
            return Optional.of(commits.get(id));
        }

        if (!isCommit(id)) {
            return Optional.empty();
        }

        return toHash(id).map(sha1 -> commits.computeIfAbsent(sha1, fullID -> new Commit(this, fullID)));
    }

    /**
     * Determines whether the given object ID designates a commit.
     *
     * @param id
     *         the ID to check
     * @return true iff the ID designates a commit
     */
    private boolean isCommit(String id) {
        Optional<ExecRes> catFile = git.exec(dir, "cat-file", "-t", id);
        Function<ExecRes, Boolean> toBoolean = res -> {

            if (git.failed(res)) {
                LOG.warning(() -> String.format("Failed to determine whether %s is a valid commit id.", id));
                return null;
            }

            boolean isCommit = res.stdOut.startsWith(TYPE_COMMIT);

            if (isCommit) {
                LOG.finer(() -> String.format("%s is a commit.", id));
            }

            return isCommit;
        };

        return catFile.map(toBoolean).orElse(false);
    }

    /**
     * Resolves the given <code>id</code> to the full SHA1 hash.
     *
     * @param id
     *         the <code>id</code> to transform
     * @return the full SHA1 hash or an empty {@link Optional} if an exception occurs
     */
    private Optional<String> toHash(String id) {
        Optional<ExecRes> revParse = git.exec(dir, "rev-parse", id);
        Function<ExecRes, String> toHash = res -> {

            if (git.failed(res)) {
                LOG.warning(() -> String.format("Failed to resolve %s to its unique SHA1 hash.", id));
                return null;
            }

            LOG.finer(() -> String.format("Resolved %s to %s.", id, res.stdOut));

            return res.stdOut;
        };

        return revParse.map(toHash);
    }

    /**
     * Copies this {@link Repository} to the given <code>destination</code> directory.
     *
     * @param destination
     *         the destination directory
     * @return a copy of this {@link Repository} at the new location
     * @see FileUtils#copyDirectory(File, File)
     */
    public Optional<Repository> copy(File destination) {

        try {
            if (destination.exists() && destination.isDirectory()) {
                LOG.warning(() -> String.format("%s already exists. Merging source and destination directories.", destination));
            }

            FileUtils.copyDirectory(dir, destination);

            LOG.fine(() -> String.format("Copied %s to %s.", this, destination));
        } catch (IOException e) {
            LOG.log(Level.WARNING, e, () -> String.format("Failed to copy %s to %s.", dir, destination));
            return Optional.empty();
        }

        Repository repo = new Repository(git, url, destination);

        for (Map.Entry<String, Commit> entry : commits.entrySet()) {
            repo.getCommitUnchecked(entry.getValue().getId());
        }

        return Optional.of(repo);
    }

    /**
     * Creates a list of the lines of a file and their corresponding commit.
     *
     * @param file the file to analyze
     * @return the list of lines and the last commit to change it
     */
    public Optional<List<BlameLine>> blameFile(Path file) {
        Path path = dir.toPath().relativize(dir.toPath().resolve(file));
        Optional<ExecRes> output = git.exec(dir, "blame", "-p", path.toString());
        Function<ExecRes, List<BlameLine>> toList = execRes -> {

            if (git.failed(execRes)) {
                LOG.warning(() -> String.format("Failed to get blame information for file %s", path));
                return null;
            }

            List<BlameLine> bLines = null;

            try {
                bLines = parseBlameLines(execRes.stdOut);
            } catch (NoSuchElementException e) {
                LOG.log(Level.WARNING, e, () -> String.format("Failed to parse the blame information for file %s", path));
            }

            return bLines;
        };

        return output.map(toList);
    }

    private List<BlameLine> parseBlameLines(String blameOutput) {
        Scanner lines = new Scanner(blameOutput);

        if (!lines.hasNextLine()) {
            return Collections.emptyList();
        }

        List<BlameLine> bLines = new ArrayList<>();
        Map<Commit, BlameLine> firstOccurrences = new HashMap<>();

        while (lines.hasNextLine()) {
            String line;
            Scanner lineScanner;

            // Read the first line of the commit header and construct/get the Commit object.
            line = lines.nextLine();
            lineScanner = new Scanner(line);

            String sha1 = lineScanner.next();
            int originalLineNumber = lineScanner.nextInt();
            int finalLineNumber = lineScanner.nextInt();

            Commit commit = getCommitUnchecked(sha1);

            Instant authorInstant = null;
            Instant committerInstant = null;
            ZoneId authorTZ = null;
            ZoneId committerTZ = null;

            File commitFile = null;

            Map<String, String> other = new HashMap<>();

            // Read the rest of the commit header and input all fields into the Commit object or the BlameLine fields.
            while (!(line = lines.nextLine()).startsWith("\t")) {
                lineScanner = new Scanner(line);
                String headerKey = lineScanner.next();

                switch (headerKey) {
                    case AUTHOR:
                        commit.setAuthor(lineScanner.nextLine().trim());
                        break;
                    case AUTHOR_MAIL: {
                        String email = lineScanner.nextLine().trim();
                        commit.setAuthorMail(email.substring(1, email.length() - 1));
                    } break;
                    case AUTHOR_TIME:
                        authorInstant = Instant.ofEpochSecond(lineScanner.nextLong());

                        if (authorTZ != null) {
                            commit.setAuthorTime(OffsetDateTime.ofInstant(authorInstant, authorTZ));
                        }
                        break;
                    case AUTHOR_TZ:
                        authorTZ = ZoneId.of(lineScanner.next());

                        if (authorInstant != null) {
                            commit.setAuthorTime(OffsetDateTime.ofInstant(authorInstant, authorTZ));
                        }
                        break;
                    case COMMITTER:
                        commit.setCommitter(lineScanner.nextLine().trim());
                        break;
                    case COMMITTER_MAIL: {
                        String email = lineScanner.nextLine().trim();
                        commit.setCommitterMail(email.substring(1, email.length() - 1));
                    } break;
                    case COMMITTER_TIME:
                        committerInstant = Instant.ofEpochSecond(lineScanner.nextLong());

                        if (committerTZ != null) {
                            commit.setCommitterTime(OffsetDateTime.ofInstant(committerInstant, committerTZ));
                        }
                        break;
                    case COMMITTER_TZ:
                        committerTZ = ZoneId.of(lineScanner.next());

                        if (committerInstant != null) {
                            commit.setCommitterTime(OffsetDateTime.ofInstant(committerInstant, committerTZ));
                        }
                        break;
                    case FILENAME:
                        commitFile = dir.toPath().resolve(Paths.get(lineScanner.nextLine().trim())).toFile();
                        break;
                    default:
                        other.put(headerKey, lineScanner.nextLine().trim());
                }
            }

            // Read the line starting with a TAB and store it in a BlameLine along with its commit.
            BlameLine bLine;
            line = line.substring(1); // Remove the TAB.

            if (firstOccurrences.containsKey(commit)) {
                bLine = BlameLine.nextOccurrence(firstOccurrences.get(commit), originalLineNumber, finalLineNumber, line);
            } else {
                bLine = BlameLine.firstOccurrence(commit, originalLineNumber, finalLineNumber, commitFile, line, other);
                firstOccurrences.put(commit, bLine);
            }

            bLines.add(bLine);
        }

        return bLines;
    }

    /**
     * Gets a list of merge conflicts fo a file, as a pair of right side/left side of list of pairs of lines with their
     * commits.
     *
     * @param file the file to analyze
     * @return list of failed chunks
     */
    public Optional<List<Pair<List<BlameLine>, List<BlameLine>>>> blameMergeConflicts(Path file) {
        Optional<List<BlameLine>> blame = blameFile(file);

        if (!blame.isPresent()) {
            return Optional.empty();
        }

        List<Pair<List<BlameLine>, List<BlameLine>>> conflicts = new ArrayList<>();

        Pair<List<BlameLine>, List<BlameLine>> currentConflict = null;
        String lastMarker = null;
        for (BlameLine line : blame.get()) {
            if (line.line.startsWith(CONFLICT_START)) {
                currentConflict = new Pair<>(new ArrayList<>(), new ArrayList<>());
                lastMarker = CONFLICT_START;
            } else if (line.line.startsWith(CONFLICT_MIDDLE)) {
                lastMarker = CONFLICT_MIDDLE;
            } else if (line.line.startsWith(CONFLICT_END)) {
                conflicts.add(currentConflict);
                lastMarker = CONFLICT_END;
            } else if (Objects.equals(lastMarker, CONFLICT_START)) {
                //noinspection ConstantConditions: there is always a start marker before a relevant line.
                currentConflict.getKey().add(line);
            } else if (Objects.equals(lastMarker, CONFLICT_MIDDLE)) {
                //noinspection ConstantConditions: there is always a start marker before a relevant line.
                currentConflict.getValue().add(line);
            }
        }

        return Optional.of(conflicts);
    }

    /**
     * Returns the {@link GitWrapper} used by this {@link Repository}.
     *
     * @return the {@link GitWrapper}
     */
    GitWrapper getGit() {
        return git;
    }

    /**
     * Returns the URL this {@link Repository} was cloned from.
     *
     * @return the URL
     */
    public String getUrl() {
        return url;
    }

    /**
     * Returns the directory containing this {@link Repository}.
     *
     * @return the repository directory
     */
    public File getDir() {
        return dir;
    }

    /**
     * Returns the name of the directory this {@link Repository} is based in.
     *
     * @return the name of the {@link Repository}
     */
    public String getName() {
        return getDir().getName();
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }

        if (o == null || getClass() != o.getClass()) {
            return false;
        }

        Repository that = (Repository) o;
        return Objects.equals(url, that.url) && Objects.equals(dir, that.dir);
    }

    @Override
    public int hashCode() {
        return Objects.hash(url, dir);
    }

    @Override
    public String toString() {
        return String.format("Repository{url=%s, dir=%s}", url, dir);
    }
}
